@@ -2,7 +2,7 @@
  -
  - https://git-annex.branchable.com/design/p2p_protocol_over_http/
  -
- - Copyright 2024 Joey Hess <id@joeyh.name>
+ - Copyright 2024-2025 Joey Hess <id@joeyh.name>
  -
  - Licensed under the GNU AGPL version 3 or higher.
  -}
@@ -10,6 +10,7 @@
 {-# LANGUAGE BangPatterns #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE RankNTypes #-}
 {-# LANGUAGE TupleSections #-}
 
 module P2P.Http.State where
@@ -44,8 +45,6 @@ import qualified Data.Map.Strict as M
 import qualified Data.Set as S
 import Control.Concurrent.Async
 import Data.Time.Clock.POSIX
-import qualified Data.Semigroup as Sem
-import Prelude
 
 data P2PHttpServerState = P2PHttpServerState
 	{ servedRepos :: M.Map UUID PerRepoServerState
@@ -62,7 +61,7 @@ instance Monoid P2PHttpServerState where
 		, updateRepos = const mempty
 		}
 
-instance Sem.Semigroup P2PHttpServerState where
+instance Semigroup P2PHttpServerState where
 	a <> b = P2PHttpServerState
 		{ servedRepos = servedRepos a <> servedRepos b
 		, serverShutdownCleanup = do
@@ -77,8 +76,11 @@ instance Sem.Semigroup P2PHttpServerState where
 data PerRepoServerState = PerRepoServerState
 	{ acquireP2PConnection :: AcquireP2PConnection
 	, annexWorkerPool :: AnnexWorkerPool
+	, annexState :: TMVar Annex.AnnexState
+	, annexRead :: Annex.AnnexRead
 	, getServerMode :: GetServerMode
 	, openLocks :: TMVar (M.Map LockID Locker)
+	, lockedFilesQSem :: LockedFilesQSem
 	}
 
 type AnnexWorkerPool = TMVar (WorkerPool (Annex.AnnexState, Annex.AnnexRead))
@@ -93,12 +95,15 @@ data ServerMode
 		}
 	| CannotServeRequests
 
-mkPerRepoServerState :: AcquireP2PConnection -> AnnexWorkerPool -> GetServerMode -> IO PerRepoServerState
-mkPerRepoServerState acquireconn annexworkerpool getservermode = PerRepoServerState
+mkPerRepoServerState :: AcquireP2PConnection -> AnnexWorkerPool -> TMVar Annex.AnnexState -> Annex.AnnexRead -> GetServerMode -> LockedFilesQSem -> IO PerRepoServerState
+mkPerRepoServerState acquireconn annexworkerpool annexstate annexread getservermode lockedfilesqsem = PerRepoServerState
 	<$> pure acquireconn
 	<*> pure annexworkerpool
+	<*> pure annexstate
+	<*> pure annexread
 	<*> pure getservermode
 	<*> newTMVarIO mempty
+	<*> pure lockedfilesqsem
 
 data ActionClass = ReadAction | WriteAction | RemoveAction | LockAction
 	deriving (Eq)
@@ -106,6 +111,7 @@ data ActionClass = ReadAction | WriteAction | RemoveAction | LockAction
 withP2PConnection
 	:: APIVersion v
 	=> v
+	-> AnnexActionRunner
 	-> TMVar P2PHttpServerState
 	-> B64UUID ClientSide
 	-> B64UUID ServerSide
@@ -116,8 +122,8 @@ withP2PConnection
 	-> (ConnectionParams -> ConnectionParams)
 	-> ((P2PConnectionPair, PerRepoServerState) -> Handler (Either ProtoFailure a))
 	-> Handler a
-withP2PConnection apiver mst cu su bypass sec auth actionclass fconnparams connaction =
-	withP2PConnection' apiver mst cu su bypass sec auth actionclass fconnparams connaction'
+withP2PConnection apiver runner mst cu su bypass sec auth actionclass fconnparams connaction =
+	withP2PConnection' apiver runner mst cu su bypass sec auth actionclass fconnparams connaction'
   where
 	connaction' conn = connaction conn >>= \case
 		Right r -> return r
@@ -127,6 +133,7 @@ withP2PConnection apiver mst cu su bypass sec auth actionclass fconnparams conna
 withP2PConnection'
 	:: APIVersion v
 	=> v
+	-> AnnexActionRunner
 	-> TMVar P2PHttpServerState
 	-> B64UUID ClientSide
 	-> B64UUID ServerSide
@@ -137,14 +144,15 @@ withP2PConnection'
 	-> (ConnectionParams -> ConnectionParams)
 	-> ((P2PConnectionPair, PerRepoServerState) -> Handler a)
 	-> Handler a
-withP2PConnection' apiver mst cu su bypass sec auth actionclass fconnparams connaction = do
-	(conn, st) <- getP2PConnection apiver mst cu su bypass sec auth actionclass fconnparams
+withP2PConnection' apiver runner mst cu su bypass sec auth actionclass fconnparams connaction = do
+	(conn, st) <- getP2PConnection apiver runner mst cu su bypass sec auth actionclass fconnparams
 	connaction (conn, st)
 		`finally` liftIO (releaseP2PConnection conn)
 
 getP2PConnection
 	:: APIVersion v
 	=> v
+	-> AnnexActionRunner
 	-> TMVar P2PHttpServerState
 	-> B64UUID ClientSide
 	-> B64UUID ServerSide
@@ -154,24 +162,37 @@ getP2PConnection
 	-> ActionClass
 	-> (ConnectionParams -> ConnectionParams)
 	-> Handler (P2PConnectionPair, PerRepoServerState)
-getP2PConnection apiver mst cu su bypass sec auth actionclass fconnparams =
-	checkAuthActionClass mst su sec auth actionclass go
-  where
-	go st servermode = liftIO (acquireP2PConnection st cp) >>= \case
+getP2PConnection apiver runner mst cu su bypass sec auth actionclass fconnparams =
+	checkAuthActionClass mst su sec auth actionclass $
+		getP2PConnection' apiver runner cu su bypass fconnparams
+
+getP2PConnection'
+	:: APIVersion v
+	=> v
+	-> AnnexActionRunner
+	-> B64UUID ClientSide
+	-> B64UUID ServerSide
+	-> [B64UUID Bypass]
+	-> (ConnectionParams -> ConnectionParams)
+	-> PerRepoServerState
+	-> P2P.ServerMode
+	-> Handler (P2PConnectionPair, PerRepoServerState)
+getP2PConnection' apiver runner cu su bypass fconnparams st servermode =
+	liftIO (acquireP2PConnection st runner cp) >>= \case
 		Left (ConnectionFailed err) -> 
 			throwError err502 { errBody = encodeBL err }
 		Left TooManyConnections ->
 			throwError err503
 		Right v -> return (v, st)
-	  where
-		cp = fconnparams $ ConnectionParams
-			{ connectionProtocolVersion = protocolVersion apiver
-			, connectionServerUUID = fromB64UUID su
-			, connectionClientUUID = fromB64UUID cu
-			, connectionBypass = map fromB64UUID bypass
-			, connectionServerMode = servermode
-			, connectionWaitVar = True
-			}
+  where
+	cp = fconnparams $ ConnectionParams
+		{ connectionProtocolVersion = protocolVersion apiver
+		, connectionServerUUID = fromB64UUID su
+		, connectionClientUUID = fromB64UUID cu
+		, connectionBypass = map fromB64UUID bypass
+		, connectionServerMode = servermode
+		, connectionWaitVar = True
+		}
 
 getPerRepoServerState :: TMVar P2PHttpServerState -> B64UUID ServerSide -> IO (Maybe PerRepoServerState)
 getPerRepoServerState mstv su = do
@@ -253,17 +274,40 @@ proxyClientNetProto conn = runNetProto
 	(clientRunState conn) (clientP2PConnection conn)
 
 type AcquireP2PConnection
-	= ConnectionParams
+	= AnnexActionRunner
+	-> ConnectionParams
 	-> IO (Either ConnectionProblem P2PConnectionPair)
 
+type LockedFilesQSem = TMVar Integer
+
+mkLockedFilesQSem :: Maybe Integer -> IO LockedFilesQSem
+mkLockedFilesQSem = newTMVarIO . fromMaybe 100
+
+consumeLockedFilesQSem :: PerRepoServerState -> IO Bool
+consumeLockedFilesQSem st = atomically $ do
+	n <- takeTMVar (lockedFilesQSem st)
+	if n < 1
+		then do
+			putTMVar (lockedFilesQSem st) n
+			return False
+		else do
+			putTMVar (lockedFilesQSem st) (pred n)
+			return True
+
+releaseLockedFilesQSem :: PerRepoServerState -> IO ()
+releaseLockedFilesQSem st = atomically $ do
+	n <- takeTMVar (lockedFilesQSem st)
+	putTMVar (lockedFilesQSem st) (succ n)
+
 mkP2PHttpServerState
 	:: GetServerMode
 	-> UpdateRepos
 	-> ProxyConnectionPoolSize
 	-> ClusterConcurrency
+	-> LockedFilesQSem
 	-> AnnexWorkerPool
 	-> Annex P2PHttpServerState
-mkP2PHttpServerState getservermode updaterepos proxyconnectionpoolsize clusterconcurrency workerpool = do
+mkP2PHttpServerState getservermode updaterepos proxyconnectionpoolsize clusterconcurrency lockedfilesqsem workerpool = do
 	enableInteractiveBranchAccess
 	myuuid <- getUUID
 	myproxies <- M.lookup myuuid <$> getProxies
@@ -277,16 +321,22 @@ mkP2PHttpServerState getservermode updaterepos proxyconnectionpoolsize clusterco
 		liftIO $ atomically $ putTMVar endv ()
 		liftIO $ wait asyncservicer
 	let servinguuids = myuuid : map proxyRemoteUUID (maybe [] S.toList myproxies)
-	st <- liftIO $ mkPerRepoServerState (acquireconn reqv) workerpool getservermode
+	annexstate <- liftIO . newTMVarIO =<< dupState
+	annexread <- Annex.getRead id
+	st <- liftIO $ mkPerRepoServerState 
+		(acquireconn reqv annexstate annexread)
+		workerpool annexstate annexread getservermode lockedfilesqsem
 	return $ P2PHttpServerState
 		{ servedRepos = M.fromList $ zip servinguuids (repeat st)
 		, serverShutdownCleanup = endit
 		, updateRepos = updaterepos
 		}
   where
-	acquireconn reqv connparams = do
+	acquireconn reqv annexstate annexread runnertype connparams = do
+		ready <- newEmptyTMVarIO
 		respvar <- newEmptyTMVarIO
-		atomically $ putTMVar reqv (connparams, respvar)
+		atomically $ putTMVar reqv (runnertype, annexstate, annexread, connparams, ready, respvar)
+		() <- atomically $ takeTMVar ready
 		atomically $ takeTMVar respvar
 
 	servicer myuuid myproxies proxypool reqv relv endv = do
@@ -298,8 +348,8 @@ mkP2PHttpServerState getservermode updaterepos proxyconnectionpoolsize clusterco
 					`orElse` 
 				(Left . Left <$> takeTMVar endv)
 		case reqrel of
-			Right (connparams, respvar) -> do
-				servicereq myuuid myproxies proxypool relv connparams
+			Right (runnertype, annexstate, annexread, connparams, ready, respvar) -> do
+				servicereq runnertype annexstate annexread myuuid myproxies proxypool relv connparams ready
 					>>= atomically . putTMVar respvar
 				servicer myuuid myproxies proxypool reqv relv endv
 			Left (Right releaseconn) -> do
@@ -307,36 +357,40 @@ mkP2PHttpServerState getservermode updaterepos proxyconnectionpoolsize clusterco
 				servicer myuuid myproxies proxypool reqv relv endv
 			Left (Left ()) -> return ()
 	
-	servicereq myuuid myproxies proxypool relv connparams
+	servicereq runnertype annexstate annexread myuuid myproxies proxypool relv connparams ready
 		| connectionServerUUID connparams == myuuid =
-			localConnection relv connparams workerpool
+			localConnection relv connparams runner ready
 		| otherwise =
 			atomically (getProxyConnectionPool proxypool connparams) >>= \case
-				Just conn -> proxyConnection proxyconnectionpoolsize relv connparams workerpool proxypool conn
-				Nothing -> checkcanproxy myproxies proxypool relv connparams
+				Just conn -> 
+					proxyConnection proxyconnectionpoolsize relv connparams runner proxypool conn ready
+				Nothing -> checkcanproxy runnertype annexstate annexread myproxies proxypool relv connparams ready
+	  where
+		runner = annexActionRunner runnertype workerpool annexstate annexread
 
-	checkcanproxy myproxies proxypool relv connparams = 
-		inAnnexWorker' workerpool
+	checkcanproxy runnertype annexstate annexread myproxies proxypool relv connparams ready =
+		runner
 			(checkCanProxy' myproxies (connectionServerUUID connparams))
 		>>= \case
 			Right (Left reason) -> return $ Left $
 				ConnectionFailed $ 
 					fromMaybe "unknown uuid" reason
 			Right (Right (Right proxyremote)) -> proxyconnection $
-				openProxyConnectionToRemote workerpool
+				runner $ openProxyConnectionToRemote
 					(connectionProtocolVersion connparams)
 					bypass proxyremote
 			Right (Right (Left clusteruuid)) -> proxyconnection $
-				openProxyConnectionToCluster workerpool
+				runner $ openProxyConnectionToCluster
 					(connectionProtocolVersion connparams)
 					bypass clusteruuid clusterconcurrency
 			Left ex -> return $ Left $
 				ConnectionFailed $ show ex
 	  where
+		runner = annexActionRunner runnertype workerpool annexstate annexread
 		bypass = P2P.Bypass $ S.fromList $ connectionBypass connparams
 		proxyconnection openconn = openconn >>= \case
 			Right conn -> proxyConnection proxyconnectionpoolsize
-				relv connparams workerpool proxypool conn
+				relv connparams runner proxypool conn ready
 			Left ex -> return $ Left $
 				ConnectionFailed $ show ex
 
@@ -355,11 +409,13 @@ data P2PConnectionPair = P2PConnectionPair
 localConnection
 	:: TMVar (IO ())
 	-> ConnectionParams
-	-> AnnexWorkerPool
+	-> (Annex () -> IO (Either SomeException ()))
+	-> TMVar ()
 	-> IO (Either ConnectionProblem P2PConnectionPair)
-localConnection relv connparams workerpool = 
+localConnection relv connparams runner ready = 
 	localP2PConnectionPair connparams relv $ \serverrunst serverconn ->
-		inAnnexWorker' workerpool $
+		runner $ do
+			liftIO $ atomically $ putTMVar ready ()
 			void $ runFullProto serverrunst serverconn $
 				P2P.serveOneCommandAuthed
 					(connectionServerMode connparams)
@@ -430,36 +486,37 @@ proxyConnection
 	:: ProxyConnectionPoolSize
 	-> TMVar (IO ())
 	-> ConnectionParams
-	-> AnnexWorkerPool
+	-> (Annex () -> IO (Either SomeException ()))
 	-> TMVar ProxyConnectionPool
 	-> ProxyConnection
+	-> TMVar ()
 	-> IO (Either ConnectionProblem P2PConnectionPair)
-proxyConnection proxyconnectionpoolsize relv connparams workerpool proxypool proxyconn = do
+proxyConnection proxyconnectionpoolsize relv connparams runner proxypool proxyconn ready = do
 	(clientconn, proxyfromclientconn) <- 
 		mkP2PConnectionPair connparams ("http client", "proxy")
 	clientrunst <- mkClientRunState connparams
 	proxyfromclientrunst <- mkClientRunState connparams
-	asyncworker <- async $
-		inAnnexWorker' workerpool $ do
-			proxystate <- liftIO Proxy.mkProxyState
-			let proxyparams = Proxy.ProxyParams
-				{ Proxy.proxyMethods = mkProxyMethods
-				, Proxy.proxyState = proxystate
-				, Proxy.proxyServerMode = connectionServerMode connparams
-				, Proxy.proxyClientSide = Proxy.ClientSide proxyfromclientrunst proxyfromclientconn
-				, Proxy.proxyUUID = proxyConnectionRemoteUUID proxyconn
-				, Proxy.proxySelector = proxyConnectionSelector proxyconn
-				, Proxy.proxyConcurrencyConfig = proxyConnectionConcurrency proxyconn
-				, Proxy.proxyClientProtocolVersion = connectionProtocolVersion connparams
-				}
-			let proxy mrequestmessage = case mrequestmessage of
-				Just requestmessage -> do
-					Proxy.proxyRequest proxydone proxyparams
-						requestcomplete requestmessage protoerrhandler
-				Nothing -> return ()
-			protoerrhandler proxy $
-				liftIO $ runNetProto proxyfromclientrunst proxyfromclientconn $
-					P2P.net P2P.receiveMessage
+	asyncworker <- async $ runner $ do
+		liftIO $ atomically $ putTMVar ready ()
+		proxystate <- liftIO Proxy.mkProxyState
+		let proxyparams = Proxy.ProxyParams
+			{ Proxy.proxyMethods = mkProxyMethods
+			, Proxy.proxyState = proxystate
+			, Proxy.proxyServerMode = connectionServerMode connparams
+			, Proxy.proxyClientSide = Proxy.ClientSide proxyfromclientrunst proxyfromclientconn
+			, Proxy.proxyUUID = proxyConnectionRemoteUUID proxyconn
+			, Proxy.proxySelector = proxyConnectionSelector proxyconn
+			, Proxy.proxyConcurrencyConfig = proxyConnectionConcurrency proxyconn
+			, Proxy.proxyClientProtocolVersion = connectionProtocolVersion connparams
+			}
+		let proxy mrequestmessage = case mrequestmessage of
+			Just requestmessage -> do
+				Proxy.proxyRequest proxydone proxyparams
+					requestcomplete requestmessage protoerrhandler
+			Nothing -> return ()
+		protoerrhandler proxy $
+			liftIO $ runNetProto proxyfromclientrunst proxyfromclientconn $
+				P2P.net P2P.receiveMessage
 	
 	let closebothsides = do
 		liftIO $ closeConnection proxyfromclientconn
@@ -497,8 +554,7 @@ proxyConnection proxyconnectionpoolsize relv connparams workerpool proxypool pro
 	
 	requestcomplete () = return ()
 	
-	closeproxyconnection = 
-		void . inAnnexWorker' workerpool . proxyConnectionCloser
+	closeproxyconnection = void . runner . proxyConnectionCloser
 
 data Locker = Locker
 	{ lockerThread :: Async ()
@@ -515,6 +571,13 @@ mkLocker :: (IO (Maybe a)) -> (a -> IO ()) -> IO (Maybe (Locker, LockID))
 mkLocker lock unlock = do
 	lv <- newEmptyTMVarIO
 	timeoutdisablev <- newEmptyTMVarIO
+	timeouttid <- async $ whenM (atomically $ readTMVar lv) $ do
+		threadDelaySeconds $ Seconds $ fromIntegral $
+			durationSeconds p2pDefaultLockContentRetentionDuration
+		atomically (tryReadTMVar timeoutdisablev) >>= \case
+			Nothing -> void $ atomically $
+				writeTMVar lv False
+			Just () -> noop
 	let setlocked = putTMVar lv
 	locktid <- async $ lock >>= \case
 		Nothing ->
@@ -530,13 +593,6 @@ mkLocker lock unlock = do
 	locksuccess <- atomically $ readTMVar lv
 	if locksuccess
 		then do
-			timeouttid <- async $ do
-				threadDelaySeconds $ Seconds $ fromIntegral $
-					durationSeconds p2pDefaultLockContentRetentionDuration
-				atomically (tryReadTMVar timeoutdisablev) >>= \case
-					Nothing -> void $ atomically $
-						writeTMVar lv False
-					Just () -> noop
 			tid <- async $ do
 				wait locktid
 				cancel timeouttid
@@ -587,11 +643,8 @@ withAnnexWorkerPool mc a = do
 			Nothing -> giveup "Use -Jn or set annex.jobs to configure the number of worker threads."
 			Just wp -> a wp
 
-inAnnexWorker :: PerRepoServerState -> Annex a -> IO (Either SomeException a)
-inAnnexWorker st = inAnnexWorker' (annexWorkerPool st)
-
-inAnnexWorker' :: AnnexWorkerPool -> Annex a -> IO (Either SomeException a)
-inAnnexWorker' poolv annexaction = do
+inAnnexWorker :: AnnexWorkerPool -> Annex a -> IO (Either SomeException a)
+inAnnexWorker poolv annexaction = do
 	(workerstrd, workerstage) <- atomically $ waitStartWorkerSlot poolv
 	resv <- newEmptyTMVarIO
 	aid <- async $ do
@@ -613,6 +666,34 @@ inAnnexWorker' poolv annexaction = do
 		putTMVar poolv pool'
 	return res
 
+handleRequestAnnex :: PerRepoServerState -> Annex a -> IO (Either SomeException a)
+handleRequestAnnex st = handleRequestAnnex' (annexState st) (annexRead st)
+
+handleRequestAnnex' :: TMVar Annex.AnnexState -> Annex.AnnexRead -> Annex a -> IO (Either SomeException a)
+handleRequestAnnex' stv rd a = do
+	ast <- atomically $ readTMVar stv
+	let astdup = dupState' ast
+
+	(res, (astdup', _)) <- Annex.run (astdup, rd) $ 
+		tryNonAsync a
+	
+	ast' <- atomically $ takeTMVar stv
+	((), (ast'', _)) <- Annex.run (ast', rd) $ mergeState astdup'
+	atomically $ putTMVar stv ast''
+
+	return res
+
+data AnnexActionRunner
+	-- Number of concurrent Annex actions is bounded by the size of the 
+	-- worker pool
+	= WorkerPoolRunner
+	-- When using this, must enforce bounds first
+	| RequestRunner
+
+annexActionRunner :: AnnexActionRunner -> AnnexWorkerPool -> TMVar Annex.AnnexState -> Annex.AnnexRead -> Annex a -> IO (Either SomeException a)
+annexActionRunner WorkerPoolRunner workerpool _ _ = inAnnexWorker workerpool
+annexActionRunner RequestRunner _ st rd = handleRequestAnnex' st rd
+
 data ProxyConnection = ProxyConnection
 	{ proxyConnectionRemoteUUID :: UUID
 	, proxyConnectionSelector :: Proxy.ProxySelector
@@ -646,38 +727,34 @@ openedProxyConnection u desc selector closer concurrency = do
 		fastDebug "P2P.Http" ("Closed proxy connection to " ++ desc)
 
 openProxyConnectionToRemote
-	:: AnnexWorkerPool
-	-> P2P.ProtocolVersion
+	:: P2P.ProtocolVersion
 	-> P2P.Bypass
 	-> Remote
-	-> IO (Either SomeException ProxyConnection)
-openProxyConnectionToRemote workerpool clientmaxversion bypass remote =
-	inAnnexWorker' workerpool $ do
-		remoteside <- proxyRemoteSide clientmaxversion bypass remote
-		concurrencyconfig <- Proxy.noConcurrencyConfig
-		openedProxyConnection (Remote.uuid remote)
-			("remote " ++ Remote.name remote)
-			(Proxy.singleProxySelector remoteside)
-			(Proxy.closeRemoteSide remoteside)
-			concurrencyconfig
+	-> Annex ProxyConnection
+openProxyConnectionToRemote clientmaxversion bypass remote = do
+	remoteside <- proxyRemoteSide clientmaxversion bypass remote
+	concurrencyconfig <- Proxy.noConcurrencyConfig
+	openedProxyConnection (Remote.uuid remote)
+		("remote " ++ Remote.name remote)
+		(Proxy.singleProxySelector remoteside)
+		(Proxy.closeRemoteSide remoteside)
+		concurrencyconfig
 
 type ClusterConcurrency = Int
 
 openProxyConnectionToCluster
-	:: AnnexWorkerPool
-	-> P2P.ProtocolVersion
+	:: P2P.ProtocolVersion
 	-> P2P.Bypass
 	-> ClusterUUID
 	-> ClusterConcurrency
-	-> IO (Either SomeException ProxyConnection)
-openProxyConnectionToCluster workerpool clientmaxversion bypass clusteruuid concurrency =
-	inAnnexWorker' workerpool $ do
-		(proxyselector, closenodes) <-
-			clusterProxySelector clusteruuid clientmaxversion bypass
-		concurrencyconfig <- Proxy.mkConcurrencyConfig concurrency
-		openedProxyConnection (fromClusterUUID clusteruuid)
-			("cluster " ++ fromUUID (fromClusterUUID clusteruuid))
-			proxyselector closenodes concurrencyconfig
+	-> Annex ProxyConnection
+openProxyConnectionToCluster clientmaxversion bypass clusteruuid concurrency = do
+	(proxyselector, closenodes) <-
+		clusterProxySelector clusteruuid clientmaxversion bypass
+	concurrencyconfig <- Proxy.mkConcurrencyConfig concurrency
+	openedProxyConnection (fromClusterUUID clusteruuid)
+		("cluster " ++ fromUUID (fromClusterUUID clusteruuid))
+		proxyselector closenodes concurrencyconfig
 
 type ProxyConnectionPool = (Integer, M.Map ProxyConnectionPoolKey [ProxyConnection])
 
